import torch
from torch import nn
import torch.nn.init as init
# from torchvision import models


class ConvBlock(nn.Module):
    def __init__(self):
        super(ConvBlock, self).__init__()
        self.conv = nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3, stride=1, padding=1, bias=True)
        self.prelu = nn.PReLU()

    def forward(self, x):
        return self.prelu(self.conv(x))


class ResBlock(nn.Module):
    def __init__(self):
        super(ResBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_channels=64, out_channels=128, kernel_size=3, stride=1, padding=1, bias=True)
        self.prelu = nn.PReLU()
        self.conv2 = nn.Conv2d(in_channels=128, out_channels=64, kernel_size=3, stride=1, padding=1, bias=True)
        self._initialize_weights()

    def forward(self, x):
        x_ = x
        out1 = self.conv2(self.prelu(self.conv1(x)))
        out = torch.add(x_, out1)
        return out

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                init.kaiming_normal_(m.weight)
                if m.bias is not None:
                    init.constant_(m.bias, 0)


class CALayer(nn.Module):
    def __init__(self, channels=64, ratio=16):
        super(CALayer, self).__init__()
        self.cal_conv1 = nn.Sequential(
            nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3,  stride=1, padding=1, bias=True), nn.PReLU())
        self.cal_conv2 = nn.Sequential(
            nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3,  stride=1, padding=1, bias=True), nn.PReLU())
        self.cal_avg_pool = nn.AdaptiveAvgPool2d(1)
        # self.cal_max_pool = nn.AdaptiveMaxPool2d(1)
        self.cal_fc1 = nn.Conv2d(channels, channels // ratio, 1, padding=0, bias=True)
        self.cal_fc2 = nn.Conv2d(channels // ratio, channels, 1, padding=0, bias=True)
        self.sigmoid = nn.Sigmoid()
        self.prelu = nn.PReLU()

    def forward(self, x):
        out1 = self.cal_conv1(x)
        cal_weight_avg = self.sigmoid(self.cal_fc2(self.prelu(self.cal_fc1(self.cal_avg_pool(out1)))))
        # max_out = self.sigmoid(self.fc2(self.prelu(self.fc1(self.max_pool(out1)))))
        out_cal_avg = x * cal_weight_avg
        # max_cha = x * max_out
        # out = self.cal_conv2(torch.cat((avg_cha, max_cha), 1))
        out = self.cal_conv2(out_cal_avg)
        return out


class SALayer(nn.Module):
    def __init__(self):
        super(SALayer, self).__init__()
        self.sal_conv1 = nn.Sequential(
            nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3,  stride=1, padding=1, bias=True), nn.PReLU())
        self.sal_conv2 = nn.Conv2d(in_channels=64, out_channels=1, kernel_size=3, padding=1, bias=True)
        self.sal_conv3 = nn.Sequential(
            nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3, stride=1, padding=1, bias=True), nn.PReLU())
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        out1 = self.sal_conv1(x)
        sal_weight = self.sigmoid(self.sal_conv2(out1))
        out = self.sal_conv3(x * sal_weight)
        return out


class ComplexBlock(nn.Module):
    def __init__(self):
        super(ComplexBlock, self).__init__()
        self.convs1 = nn.Sequential(
            nn.Conv2d(in_channels=1, out_channels=32, kernel_size=3, padding=1, bias=True), nn.PReLU())
        self.convs2 = nn.Sequential(
            nn.Conv2d(in_channels=2, out_channels=32, kernel_size=3, padding=1, bias=True), nn.PReLU())
        self.convs3 = nn.Sequential(
            nn.Conv2d(in_channels=2, out_channels=32, kernel_size=3, padding=1, bias=True), nn.PReLU())
        self.convs4 = nn.Sequential(
            nn.Conv2d(in_channels=1, out_channels=32, kernel_size=3, padding=1, bias=True), nn.PReLU())
        self.convs5 = nn.Sequential(
            nn.Conv2d(in_channels=2, out_channels=32, kernel_size=3, padding=1, bias=True), nn.PReLU())
        self.convs6 = nn.Sequential(
            nn.Conv2d(in_channels=1, out_channels=32, kernel_size=3, padding=1, bias=True), nn.PReLU())
        self.conv_cb = nn.Sequential(
            nn.Conv2d(in_channels=192, out_channels=64, kernel_size=3, padding=1, bias=True), nn.PReLU())

    def forward(self, x):
        conv1 = self.convs1(torch.index_select(x, 1, torch.LongTensor([0]).cuda()))
        conv2 = self.convs2(torch.index_select(x, 1, torch.LongTensor([1, 2]).cuda()))
        conv3 = self.convs3(torch.index_select(x, 1, torch.LongTensor([3, 4]).cuda()))
        conv4 = self.convs4(torch.index_select(x, 1, torch.LongTensor([5]).cuda()))
        conv5 = self.convs5(torch.index_select(x, 1, torch.LongTensor([6, 7]).cuda()))
        conv6 = self.convs6(torch.index_select(x, 1, torch.LongTensor([8]).cuda()))
        out = self.conv_cb(torch.cat((conv1, conv2, conv3, conv4, conv5, conv6), 1))
        return out


class LPSR(nn.Module):
    def __init__(self):
        super(LPSR, self).__init__()
        self.lpsr_cb = ComplexBlock()
        # self.lpsr_upsample = torch.nn.UpsamplingBilinear2d(scale_factor=2)
        self.lpsr_deconv = nn.Sequential(
            nn.ConvTranspose2d(in_channels=64, out_channels=64, kernel_size=4, stride=2, padding=1, bias=True), nn.PReLU())
        self.lpsr_residual = self.make_layer(ResBlock, 5)
        # self.lpsr_cal = CALayer()
        self.initialize_weights()

    def make_layer(self, block, num_of_layer):
        layers = []
        for _ in range(num_of_layer):
            layers.append(block())
        return nn.Sequential(*layers)

    def initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                init.kaiming_normal_(m.weight)
                if m.bias is not None:
                    init.constant_(m.bias, 0)

    def forward(self, x):
        out1 = self.lpsr_deconv(self.lpsr_cb(x))
        # out2 = self.lpsr_cal(self.lpsr_residual(out1))
        out2 = self.lpsr_residual(out1)
        out = torch.add(out1, out2)
        return out


class HSFE(nn.Module):
    def __init__(self):
        super(HSFE, self).__init__()
        self.hsfe_input = nn.Sequential(
            nn.Conv2d(in_channels=1, out_channels=64, kernel_size=3, padding=1, bias=True), nn.PReLU())
        self.hsfe_res = self.make_layer(ResBlock, 5)
        self.hsfe_sal = SALayer()
        self.initialize_weights()

    def make_layer(self, block, num_of_layer):
        layers = []
        for _ in range(num_of_layer):
            layers.append(block())
        return nn.Sequential(*layers)

    def initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                init.kaiming_normal_(m.weight)
                if m.bias is not None:
                    init.constant_(m.bias, 0)

    def forward(self, x):
        out1 = self.hsfe_input(x)
        out2 = self.hsfe_sal(self.hsfe_res(out1))
        out = torch.add(out1, out2)
        return out


class HRSA(nn.Module):
    def __init__(self):
        super(HRSA, self).__init__()
        self.hrsa_conv1 = nn.Sequential(
            nn.Conv2d(in_channels=1, out_channels=64, kernel_size=3,  stride=1, padding=1, bias=True), nn.PReLU())
        self.hrsa_conv2 = nn.Sequential(
            nn.Conv2d(in_channels=9, out_channels=64, kernel_size=3,  stride=1, padding=1, bias=True), nn.PReLU())
        # self.hrsa_upsample = torch.nn.UpsamplingBilinear2d(scale_factor=2)
        self.hrsa_deconv = nn.Sequential(
            nn.ConvTranspose2d(in_channels=64, out_channels=64, kernel_size=4, stride=2, padding=1, bias=True), nn.PReLU())
        self.hrsa_conv3 = nn.Conv2d(in_channels=64, out_channels=1, kernel_size=3, padding=1, bias=True)
        self.hrsa_conv4 = nn.Sequential(
            nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3,  stride=1, padding=1, bias=True), nn.PReLU())
        self.sigmoid = nn.Sigmoid()
        self.initialize_weights()

    def initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                init.kaiming_normal_(m.weight)
                if m.bias is not None:
                    init.constant_(m.bias, 0)

    def forward(self, lr, hr):
        out1 = self.hrsa_conv1(hr)
        out2 = self.hrsa_deconv(self.hrsa_conv2(lr))
        out3 = self.sigmoid(self.hrsa_conv3(out1))
        out = self.hrsa_conv4(torch.mul(out2, out3))
        return out


class LRCA(nn.Module):
    def __init__(self, channels=64, ratio=16):
        super(LRCA, self).__init__()
        # self.lrca_upsample = torch.nn.UpsamplingBilinear2d(scale_factor=2)
        self.lrca_deconv = nn.Sequential(
           nn.ConvTranspose2d(in_channels=64, out_channels=64, kernel_size=4, stride=2, padding=1, bias=True), nn.PReLU())
        self.lrca_conv1 = nn.Sequential(
            nn.Conv2d(in_channels=9, out_channels=64, kernel_size=3,  stride=1, padding=1, bias=True), nn.PReLU())
        self.lrca_conv2 = nn.Sequential(
            nn.Conv2d(in_channels=1, out_channels=64, kernel_size=3,  stride=1, padding=1, bias=True), nn.PReLU())
        # self.lrca_avg_pool = nn.AdaptiveAvgPool2d(1)
        # self.lrca_max_pool = nn.AdaptiveMaxPool2d(1)
        # self.lrca_fc1 = nn.Conv2d(channels, channels // ratio, 1, padding=0, bias=True)
        # self.lrca_fc2 = nn.Conv2d(channels // ratio, channels, 1, padding=0, bias=True)
        self.lrca_conv3 = nn.Sequential(
            nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3,  stride=1, padding=1, bias=True), nn.PReLU())
        self.sigmoid = nn.Sigmoid()
        self.prelu = nn.PReLU()
        self.initialize_weights()

    def initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                init.kaiming_normal_(m.weight)
                if m.bias is not None:
                    init.constant_(m.bias, 0)

    def forward(self, lr, hr):
        out1 = self.lrca_deconv(self.lrca_conv1(lr))
        out2 = self.lrca_conv2(hr)
        # lrca_avg_weight = self.sigmoid(self.lrca_fc2(self.prelu(self.lrca_fc1(self.lrca_avg_pool(out1)))))
        lrca_avg_weight = self.sigmoid(out1)
        # max_out = self.sigmoid(self.lrca_fc2(self.prelu(self.lrca_fc1(self.lrca_max_pool(out1)))))
        out_lrca_avg = torch.mul(out2, lrca_avg_weight)
        # max_cha = out2 * max_out
        # out = self.lrca_conv3(torch.cat((avg_cha, max_cha), 1))
        out = self.lrca_conv3(out_lrca_avg)
        return out


class PSNS(nn.Module):
    def __init__(self):
        super(PSNS, self).__init__()
        self.hsfe = HSFE()
        self.lpsr = LPSR()
        self.hla = HRSA()
        self.lha = LRCA()
        # self.fus_cal = CALayer()
        # self.fus_sal = SALayer()
        self.fus_conv1 = nn.Sequential(
            nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3, padding=1, bias=True), nn.PReLU())
        self.fus_conv2 = self.make_layer(ConvBlock, 3)
        self.fus_conv3 = nn.Sequential(
            nn.Conv2d(in_channels=128, out_channels=64, kernel_size=3, padding=1, bias=True), nn.PReLU())
        # self.fus_conv4 = nn.Sequential(
        #    nn.Conv2d(in_channels=128, out_channels=64, kernel_size=3, padding=1, bias=True), nn.PReLU())
        self.fus_conv4 = self.make_layer(ConvBlock, 3)
        self.fus_output = nn.Sequential(
            nn.Conv2d(in_channels=64, out_channels=9, kernel_size=3, padding=1, bias=True), nn.PReLU())

    def make_layer(self, block, num_of_layer):
        layers = []
        for _ in range(num_of_layer):
            layers.append(block())
        return nn.Sequential(*layers)

    def forward(self, lr, hr):
        out1 = self.fus_conv2(self.fus_conv1(torch.add(self.lpsr(lr), self.hsfe(hr))))
        out2 = self.hla(lr, hr)
        out3 = self.lha(lr, hr)
        out4 = self.fus_conv3(torch.cat((out2, out3), 1))
        out = self.fus_output(self.fus_conv4(torch.add(out1, out4)))

        return out
