import os, datetime, time
import h5py
import glob
import re
import warnings
import argparse
import numpy as np
# from spsr import SPSR
from psns20200622 import PSNS
from torch import nn
import torch.optim as optim
from torch.optim.lr_scheduler import StepLR
from dataset import *
from torch.utils.data import DataLoader
from torch.autograd import Variable
import gc
warnings.filterwarnings("ignore")


parser = argparse.ArgumentParser(description='Pytorch PSFN')
parser.add_argument('--model', default='PSFN', type=str, help='choose path of model')
parser.add_argument('--batch_size', default=128, type=int, help='batch size')
parser.add_argument('--epoch', default=201, type=int, help='number of train epoch')
parser.add_argument('--lr', default=1e-4, type=float, help='initial learning rate for Adam')
parser.add_argument('--train_data', default='data/train/train.h5', type=str, help='path of train data')
parser.add_argument('--gpu', default='0', type=str, help='gpu id')
args = parser.parse_args()

cuda = torch.cuda.is_available()
os.environ["CUDA_VISIBLE_DEVICES"] = '0'

batch_size = args.batch_size
nepoch = args.epoch
savedir = os.path.join('model', args.model)

if not os.path.exists(savedir):
    os.mkdir(savedir)

log_header = [
    'epoch',
    'iteration',
    'train/loss',
]
if not os.path.exists(os.path.join(savedir, 'log.csv')):
            with open(os.path.join(savedir, 'log.csv'), 'w') as f:
                f.write(','.join(log_header) + '\n')


def find_checkpoint(savedir):
    file_list = glob.glob(os.path.join(savedir, 'model_*.pth'))
    if file_list:
        epochs_exist = []
        for m in file_list:
            result = re.findall(".*model_(.*).pth.*", m)
            epochs_exist.append(int(result[0]))
        initial_epoch = max(epochs_exist)
    else:
        initial_epoch = 0
    return initial_epoch


def main():
    # dataset generator
    print("==> Generating data")
    hf = h5py.File(args.train_data, 'r+')
    lr_polsar = np.float32(hf['data1'])
    hr_sar = np.float32(hf['data2'])
    hr_polsar = np.float32(hf['label'])

    lr_polsar = torch.from_numpy(lr_polsar).view(-1, 9, 20, 20)
    hr_sar = torch.from_numpy(hr_sar).view(-1, 1, 40, 40)
    hr_polsar = torch.from_numpy(hr_polsar).view(-1, 9, 40, 40)

    train_set = PSFNDataset(lr_polsar, hr_sar, hr_polsar)
    train_loader = DataLoader(dataset=train_set, num_workers=8, drop_last=True, batch_size=128, shuffle=True, pin_memory=True)

    print("==> Building model")
    model = PSNS()
    criterion1 = nn.MSELoss(size_average=True)
    criterion2 = nn.MSELoss(size_average=True)

    print("==> Setting optimizer")
    optimizer = optim.Adam(model.parameters(), lr=args.lr,  weight_decay=0)
    scheduler = StepLR(optimizer, step_size=50, gamma=0.5)  # learning rates adjust

    initial_epoch = find_checkpoint(savedir=savedir)
    if initial_epoch > 0:
        print('==> Resuming by loading epoch %03d' % initial_epoch)
        model = torch.load(os.path.join(savedir, 'model_%03d.pth' % initial_epoch))

    if cuda:
        print("==> Setting GPU")
        print("===> gpu id: '{}'".format(args.gpu))
        model = model.cuda()
        criterion1 = criterion1.cuda()
        criterion2 = criterion2.cuda()

    for epoch in range(initial_epoch, nepoch):
        scheduler.step(epoch)
        epoch_loss = 0
        start_time = time.time()

        # train
        model.train()
        num_train = len(train_loader.dataset)
        for iteration, batch in enumerate(train_loader):
            lr_polsar_batch, hr_sar_batch, hr_polsar_batch = Variable(batch[0]), Variable(batch[1]), Variable(batch[2])
            if cuda:
                lr_polsar_batch = lr_polsar_batch.cuda()
                hr_sar_batch = hr_sar_batch.cuda()
                hr_polsar_batch = hr_polsar_batch.cuda()
            optimizer.zero_grad()
            out = model(lr_polsar_batch, hr_sar_batch)
            out_vv = out[:, 8, :, :].view(-1, 1, 40, 40)
            loss1 = criterion1(out, hr_polsar_batch)
            loss2 = criterion2(out_vv, hr_sar_batch)

            alpha = loss1.data / (loss1.data + loss2.data)
            beta = loss2.data / (loss1.data + loss2.data)
            loss = alpha * loss1 + beta * loss2
            epoch_loss += loss.data/num_train
            print('%4d %4d / %4d loss_sr = %2.4f' % (epoch + 1, iteration, train_set.hr_polsar.size(0)/batch_size, loss1.data))
            loss.backward()
            optimizer.step()
            with open(os.path.join(savedir, 'log.csv'), 'a') as file:
                log = [epoch, iteration] + [loss1.data.item()]
                log = map(str, log)
                file.write(','.join(log) + '\n')
        torch.save(model, os.path.join(savedir, 'model_%03d.pth' % (epoch + 1)))
        gc.collect()
        elapsed_time = time.time() - start_time
        print('epcoh = %4d , time is %4.4f s' % (epoch + 1, elapsed_time))


if __name__ == '__main__':
    main()
